;=============================================================================
; Insight, real-mode debugger for MS DOS / PC DOS / FreeDOS.
; Copyright (c) Victor M. Gamayunov, Sergey Pimenov, 1993, 96, 97, 2002.
; Modifications by Oleg O. Chukaev (2006 - 2009).
;-----------------------------------------------------------------------------
; iface_2.inc
; User interface and other procedures.
;-----------------------------------------------------------------------------
; This program is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; as published by the Free Software Foundation; either version 2
; of the License, or (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
; 02110-1301 USA
;=============================================================================


;=============================================================================
; Changelog
;-----------------------------------------------------------------------------
; 2008-11-03	0leg	Fixed Alt+L in the Dump window, added Alt+F5, Alt+U.
;			Added Up/Down to the Flags window.
; 2009-01-08	0leg	GPF on BP/SP modification: fixed.
;=============================================================================


;=============================================================================
; draw_cpu_window
;-----------------------------------------------------------------------------
draw_cpu_window:
		call	pushr
		call	no_cursor
		mov	es,[video_seg]
		xor	di,di
		cld
		mov	cx,SCR_COLS*SCR_ROWS
		test	byte [cmd_line_options],CMD_EGA
		jz	@@not_ega
		add	cx,SCR_COLS*(MAX_SCR_ROWS-SCR_ROWS)
@@not_ega:
		mov	ax,atMainWindow*100h+20h
	rep	stosw

		test	byte [large_dump],1
		jz	@@not_quit
		jmp	@@quit
@@not_quit:
;;;;		jnz	@@quit			;!!!

		mov	di,CPU_START
	drawchoice '', '='
	mov al, byte [DRAWADDRESS]
		mov	ah,atCPUborder
		mov	cx,CPU_WIDTH
	rep	stosw
	drawchoice '', '#'
	mov al, byte [DRAWADDRESS]

		stosw
	drawchoice '', '|'
	mov al, byte [DRAWADDRESS]
		mov	cx,CPU_HEIGHT

@@next_bord_line:
		add	di,(SCR_COLS-1)*2
		stosw
		loop	@@next_bord_line

		mov	di,CPU_START+(CPU_HEIGHT+1)*SCR_COLS*2
	drawchoice '', '-'
	mov al, byte [DRAWADDRESS]
		mov	cx,USED_COLS
	rep	stosw
		mov	di,CPU_START+((CPU_HEIGHT+1)*SCR_COLS+CPU_WIDTH)*2
	drawchoice '', '+'
	mov al, byte [DRAWADDRESS]
		stosw

		mov	bx,FLAG16_START
		cmp	byte [regs_mode],1
		jne	@@skip_32
		mov	bx,FLAG32_START
@@skip_32:
		lea	di,[bx-SCR_COLS*2-2]

	drawchoice '', '+'
	mov al, byte [DRAWADDRESS]
		stosw
		mov	cx,USED_COLS-1-CPU_WIDTH
	drawchoice '', '-'
	mov al, byte [DRAWADDRESS]
	rep	stosw

		lea	di,[bx+2*SCR_COLS*2-2]

	drawchoice '', '+'
	mov al, byte [DRAWADDRESS]
		stosw
		mov	cx,USED_COLS-1-CPU_WIDTH
	drawchoice '', '-'
	mov al, byte [DRAWADDRESS]
	rep	stosw

;flags names
		mov	ah,atRegName
		lea	di,[bx+2]
		mov	si,flags_str
		mov	cx,NUM_FLAGS
@@loop01:
		lodsb
		stosw
		mov	al,'F'
		stosw
		scasw
		loop	@@loop01

		mov	ax,40-TITLE_LEN/2
		mov	bh,atTitle
		mov	si,main_title
		call	write_string

		mov	ax,CPU_START+5
		mov	bh,atCPUName
		mov	si,[cpu_name]
		call	write_string

		cmp	byte [regs_mode],1
		je	@@regs_32
		call	draw_registers_16
@@exit:
		mov	si,flags_msg
		call	fill_str
@@quit:
		call	popr
		ret

@@regs_32:
		call	draw_registers_32
		jmp	@@exit
;=============================================================================
; draw_registers_16
;-----------------------------------------------------------------------------
draw_registers_16:
		mov	di,REG_START
		mov	si,rreg
		mov	ah,atRegName
		mov	cx,4

@@next_reg_line:
		push	cx
		mov	cx,3

@@next_reg_col:
		lodsb
		stosw
		lodsb
		stosw
		mov	al,'='
		stosw
		add	di,12
		add	si,6
		loop	@@next_reg_col

		add	di,SCR_COLS*2-18*3
		pop	cx
		sub	si,22
		loop	@@next_reg_line

		add	di,SCR_COLS*2
		mov	al,'I'
		stosw
		mov	al,'P'
		stosw
		mov	al,'='
		stosw
		add	di,24
		ret
;=============================================================================
; draw_registers_32
;-----------------------------------------------------------------------------
draw_registers_32:
		mov	di,REG_START
		mov	si,rreg		;1st column
		mov	bp,rreg_segs	;2nd column
		mov	ah,atRegName
		mov	cx,6

@@next_reg_32:
		call	@@display_names
		loop	@@next_reg_32

		add	di,SCR_COLS*2
		call	@@display_names
		call	@@display_names

		sub	di,(SCR_COLS-15)*2
		ret

@@display_names:
		push	di
		mov	al,'E'
		stosw
		lodsb
		stosw
		lodsb
		stosw
		mov	al,'='
		stosw
		add	di,4*2
		mov	al,':'
		stosw
		add	di,18
		xchg	si,bp
		lodsb
		stosw
		lodsb
		stosw
		mov	al,'='
		stosw
		xchg	si,bp
		pop	di
		add	di,SCR_COLS*2
		ret
;=============================================================================
; update_watch_label
;-----------------------------------------------------------------------------
update_watch_label:
		cmp	byte [large_dump],0
		jz	@@upd
		ret
@@upd:
		call	pushr

		mov	es,[video_seg]
		cld
		mov	di,WATCH_START

		cmp	word [data_watch_proc],0
		je	@@draw_line

		mov	si,[data_watch_label]
		mov	ax,atDataWatchMem*256 + ' '
		stosw

		lodsb
		stosw
		lodsb
		stosw

		mov	al,':'
		stosw

		lodsb
		stosw
		lodsb
		stosw

		mov	al,' '
		stosw
@@exit:
		call	popr
		ret

@@draw_line:

	mov ah, atCPUborder
	drawchoice '', '-'
	mov al, byte [DRAWADDRESS]
		mov	cx,7
	rep	stosw
		jmp	@@exit
;=============================================================================
; fill_str
;-----------------------------------------------------------------------------
fill_str:
@@next:
		lodsb
		cmp	al,0
		je	@@quit
		stosw
		jmp	@@next
@@quit:
		ret
;=============================================================================
; update_registers
;-----------------------------------------------------------------------------
update_registers:
		cmp	byte [large_dump],0
		jz	@@upd
		ret
@@upd:
		call	pushr
		mov	es,[video_seg]

		mov	ax,FLAG16_START+SCR_COLS*2+4
		mov	cx,NUM_16_REGS
		mov	si,reg16_pos_table
		cmp	byte [regs_mode],1
		jne	@@skip_32
		mov	ax,FLAG32_START+SCR_COLS*2+4
		mov	cl,NUM_32_REGS		;CL instead of CX
		mov	si,reg32_pos_table
@@skip_32:
		push	ax

@@next_reg:
		xor	dx,dx
		mov	dl,[si+REG_POS_X]
		shl	dx,1
		mov	al,SCR_COLS*2
		mul	byte [si+REG_POS_Y]
		add	dx,ax
		mov	di,[si+REG_POS_LINK]

		mov	ax,[di]

		mov	bh,atRegNewValue
		cmp	[di+save_cpu_regs-cpu_regs],ax
		jne	@@fill
		mov	bh,atRegValue
		cmp	byte [regs_mode],1
		jne	@@fill
		call	align_reg
		; If we go here then the low value is equal.
		;  So just branch if the high value is equal too.
	push ax
	mov ax, word [di + 2]
	cmp word [di + 2 + save_cpu_regs - cpu_regs], ax
	pop ax
		je	@@fill
		mov	bh,atRegNewValue
@@fill:
		mov	di,dx
		call	fill_word
		add	si,SIZEOF_REG_POS
		loop	@@next_reg

		pop	di
;-----------------------------------------------------------------------------
@@draw_flags:
		mov	si,0000111011010101b
		;	   ----oditsz-a-p-c
		mov	cx,NUM_FLAGS

		mov	bx,[cpu_regs+REG_FL]
		mov	bp,[save_cpu_regs+REG_FL]
@@next_flag:
@@next_bit:
		mov	ax,atFlagValue*100h+'0'
		mov	dx,ax
		shl	bp,1
		adc	dl,ch		;CH instead of 0
		shl	bx,1
		adc	al,0
		shl	si,1
		jnc	@@next_bit

		xor	dx,ax
		jz	@@display_flag
		mov	ah,atFlagNewValue
@@display_flag:
		stosw
		scasw
		scasw
		loop	@@next_flag

		call	popr
		ret
;=============================================================================
; align_reg
;-----------------------------------------------------------------------------
align_reg:
		sub	di,cpu_regs
		and	di,01111100b
		add	di,cpu_regs
		ret
;=============================================================================
; update_commands
;-----------------------------------------------------------------------------
update_commands:
		cmp	byte [large_dump],0
		jz	@@upd
		ret
@@upd:
		call	pushr
		mov	byte [cur_line_brk],0
		mov	word [cur_line_ip],0
		mov	word [cur_ip_line],-1
		mov	si,[cur_ip]
		mov	di,CMD_START
		mov	es,[video_seg]
		mov	cx,CPU_HEIGHT
		xor	bp,bp

next_cpu_line:
		xor	dx,dx
		mov	bx,[unasm_seg]
		cmp	bx,[cpu_regs+REG_CS]
		jne	@@1
		cmp	si,[cpu_regs+REG_IP]
		jne	@@1
		inc	dx
		mov	[cur_ip_line],bp
@@1:
		call	check_break
		jc	@@8
		mov	dh,1		;Breakpoint exist
		jz	@@8		;2008-03-26, JE => JZ
		mov	dh,2		;Breakpoint disabled
@@8:
		mov	bx,si
		call	unassemble
		mov	ax,cmd_normal
		cmp	dh,1
		jb	@@a
		mov	ax,cmd_breakpoint
		jz	@@a
		mov	ax,cmd_brk_disabled
@@a:
		or	dl,dl
		jz	@@2
	drawchoice 11h, '<'
	drawchoice 10h, '>'
	push word [DRAWADDRESS - 1]	; get last two bytes ( or <>)
	pop word [buffer + 4]
@@2:
		cmp	bp,[cur_line]
		jne	@@display
		cmp	byte [focus],1
		jne	@@e
		cmp	ax,cmd_breakpoint
		mov	ax,cmd_selected
		jb	@@e
		mov	byte [cur_line_brk],1
		mov	ax,cmd_brk_selected
		je	@@e
		mov	ax,cmd_brk_dis_sel
@@e:
		mov	[cur_line_ip],bx
		call	update_cmd_mem

@@display:
		call	@@displayy
		call	get_ref_addresses
		call	set_direction
		call	display_cpu_line

		add	di,SCR_COLS*2
		inc	bp

		call	display_code_marks

		loop	jmp_next_cpu_line

		test	byte [cs:edit_mode],ED_STACK+ED_DUMP	;2008-03-23
		jnz	@@quit					;2008-03-23

		add	di,10
		mov	cx,[unasm_seg]
		mov	dx,[cur_line_ip]
		call	show_address
@@quit:
		call	popr
		ret

jmp_next_cpu_line equ $
	jmp next_cpu_line

;-----------------------------------------------------------------------------
@@displayy:
		cmp	byte [focus],1
		je	@@focus
		cmp	bp,[cur_line]
		jne	@@focus
		cmp	ax,cmd_breakpoint
		jae	@@focus
		mov	ax,cmd_not_focus
@@focus:
		ret
;-----------------------------------------------------------------------------

show_address:
		mov	ax,atCurUnasmAddr*100h+' '
		stosw
show_address_2:
		mov	bh,atCurUnasmAddr
		mov	ax,cx
		call	fill_word
		mov	ax,atCurUnasmAddr*100h+':'
		stosw
		mov	ax,dx
		call	fill_word
		mov	ax,atCurUnasmAddr*100h+' '
		stosw
		ret

;=============================================================================
; display_cpu_line
;-----------------------------------------------------------------------------
display_cpu_line:
		call	pushr
		xchg	ax,bp
		mov	si,buffer
		mov	dx,[cur_x]
		mov	cx,CPU_WIDTH-1

		mov	ah,[bp]
		mov	bx,CPU_ADDR_SIZE
		call	@@display

		mov	ah,[bp+1]
		mov	bx,CPU_OPCODE_SIZE
		call	@@display

		call	@@display_command

		call	popr
		ret

@@display:
		xchg	bx,cx
@@next_char:
		lodsb
		or	dx,dx
		jz	@@write_char
		dec	dx
		jmp	@@update_cx

@@write_char:
		stosw
		dec	bx

@@update_cx:
		loop	@@next_char
		xchg	bx,cx
		ret

@@display_command:
		lodsb

		mov	ah,[bp+4]
@@write_18_or_19:
		cmp	al,18h
		je	@@write_18
		cmp	al,19h
		je	@@write_19

		mov	ah,[bp+5]
		cmp	al,1eh
		je	@@disabled_arrow
		cmp	al,1fh
		je	@@disabled_arrow

		mov	ah,[bp+2]
		cmp	al,'0'
		jb	@@write
		cmp	al,'9'
		jbe	@@digit
		cmp	al,'A'
		jb	@@write
		cmp	al,'F'
		ja	@@write
@@digit:
		mov	ah,[bp+3]

@@write:
		or	dx,dx
		jz	@@stosw
		dec	dx
		jmp	@@display_command
@@stosw:
		stosw
		loop	@@display_command
		ret

@@disabled_arrow:
		sub	al,1eh-18h
		jmp	@@write_18_or_19

@@write_18:
	drawchoice 18h, '^'
	mov al, byte [DRAWADDRESS]
	jmp @@write

@@write_19:
	drawchoice 19h, 'v'
	mov al, byte [DRAWADDRESS]
	jmp @@write

;=============================================================================
; update_cmd_mem
;-----------------------------------------------------------------------------
update_cmd_mem:
		call	pushr
		mov	es,[video_seg]
		mov	di,MEM_START
	drawchoice '', '-'
	mov al, byte [DRAWADDRESS]
		mov	ah,atCPUborder
		mov	cx,21
	rep	stosw
		cmp	byte [memory_flag],0
		je	@@exit
		call	calc_mem_address
		call	disp_mem_addr
		mov	bl,[memory_flag]
		mov	bh,0
		dec	bx
		cmp	bx,2
		ja	@@exit
		mov	ds,dx
		add	bx,bx
		call	[cs:display_mem_disp+bx]
@@exit:
		call	popr
		ret

calc_mem_disp	dw	calc_bxsi
		dw	calc_bxdi
		dw	calc_bpsi
		dw	calc_bpdi
		dw	calc_si
		dw	calc_di
		dw	calc_bp
		dw	calc_bx
		dw	calc_mem

calc_bxsi:	mov	ax,[cpu_regs+REG_BX]
		add	ax,[cpu_regs+REG_SI]
		ret

calc_bxdi:	mov	ax,[cpu_regs+REG_BX]
		add	ax,[cpu_regs+REG_DI]
		ret

calc_bpsi:	mov	ax,[cpu_regs+REG_BP]
		add	ax,[cpu_regs+REG_SI]
..@ss:		mov	dx,[cpu_regs+REG_SS]
		mov	cl,'S'
		ret

calc_bpdi:	mov	ax,[cpu_regs+REG_BP]
		add	ax,[cpu_regs+REG_DI]
		jmp	..@ss

calc_si:	mov	ax,[cpu_regs+REG_SI]
		ret

calc_di:	mov	ax,[cpu_regs+REG_DI]
		ret

calc_bx:	mov	ax,[cpu_regs+REG_BX]
		ret

calc_bp:	mov	ax,[cpu_regs+REG_BP]
		jmp	..@ss

calc_mem:	xor	ax,ax
		ret
;=============================================================================
; calc_mem_address
;-----------------------------------------------------------------------------
calc_mem_address:
		mov	bl,[mem_disp_type]
		mov	bh,0
		shl	bx,1
		mov	dx,[cpu_regs+REG_DS]
		mov	cl,'D'
		call	[calc_mem_disp+bx]
		add	ax,[mem_disp_offs]
		mov	si,ax
		mov	ax,[mem_prefix]
		or	ax,ax
		jz	@@exit

		cmp	ax,pfs
		jb	@@set_prefix
		cmp	byte [system_cpu],3
		jb	@@exit
		cmp	byte [cpu_type],a386
		jb	@@exit

@@set_prefix:
		mov     bx,ax
		mov	cl,[bx]
		sub	cl,'a'-'A'
		add	bx,cpu_regs+REG_CS-pcs
		mov	dx,[bx]
@@exit:
		ret
;=============================================================================

display_mem_disp:
		dw	disp_byte
		dw	disp_word
		dw	disp_dword
disp_byte:
		mov	bh,ah
		lodsb
		call	fill_byte
		jmp	..@rest
disp_word:
		mov	bh,ah
		cmp	si,-1
		jne	@@lodsw
		lodsb
		mov	ah,al
		lodsb
@@lodsw:
		lodsw
		call	fill_word
		jmp	..@rest
disp_dword:
		mov	bh,ah
		lodsw
		xchg	ax,cx
		lodsw
		call	fill_word
		mov	ah,bh
		mov	al,':'
		stosw
		xchg	ax,cx
		call	fill_word
..@rest:
		mov	ah,bh
		mov	al,' '
		stosb
		ret
;=============================================================================
; disp_mem_addr
;-----------------------------------------------------------------------------
disp_mem_addr:
		mov	es,[video_seg]
		mov	di,MEM_START
		mov	ah,atMemAddr
		mov	al,' '
		stosw
		mov	al,cl
		stosw
		mov	al,'S'
		stosw
		mov	al,':'
		stosw
		mov	bh,ah
		mov	ax,si
		call	fill_word
		mov	ah,bh
		mov	al,' '
		stosw
		mov	al,'='
		stosw
		mov	al,' '
		stosw
		mov	ah,atMemValue
		ret

%if _TERMINAL
update_terminal:
	rol byte [terminal_opened], 1
	jnc .ret
	call pushr

	call .clear

	mov es, word [video_seg]
	mov byte [skipintro], -1
	mov si,msg_hello	; *offset*
	call MSG		; say hello
	mov byte [skipintro], 0

	call update_proto_stat_msg.update

	rol byte [terminal_connected], 1
	jc .connected
	mov ax, 100h + 26
	add ah, byte [dump_y]
	mov si, disconnected_msg
	mov bh, 7
	call write_string
	jmp .pop

.connected:
	mov si, word [terminal_buffer.draw]
	mov di, word [terminal_buffer.end]
@@:
	cmp si, di
	je @F
	mov es, word [terminal_buffer]
	es lodsb
	mov es, word [video_seg]
	call SHOW_RX
	jmp @B

@@:
.pop:
	call popr
.ret:
	retn

.clear:
	call init_terminal_size
	shr ax, 1
	mov cx, word [rend]
	sub cx, ax
	mov es, word [video_seg]
	mov ax, word [attr]
	rep stosw
	retn


init_terminal_size:
	xor ax, ax
	mov al, byte [dump_y]
	dec ax
	mov word [helpline], ax

	xor ax, ax
	add ah, byte [dump_y]
	mov word [rposxy], ax
	mov word [asaveposxy], ax

	mov word [rstartxy], ax
	mov byte [rfirstrow], ah
	call get_addr
	mov word [rstart], ax
	retn
%endif
;-----------------------------------------------------------------------------
expand_dump:
		mov	word [dump_start],SCR_COLS*2	;1st row
		mov	ax,[screen_height]
		mov	[dump_height],ax
		mov	byte [dump_y],1
		mov	byte [large_dump],1
		ret
;-----------------------------------------------------------------------------
shrink_dump:
		mov	word [dump_start],DUMP_START
		mov	ax,[e_dump_height]
		mov	[dump_height],ax
		mov	byte [dump_y],DUMP_Y
		mov	byte [large_dump],0
		ret
;-----------------------------------------------------------------------------

zoom_dump:
		test	byte [large_dump],1
		jz	@@enlarge

%if _TERMINAL
	rol byte [terminal_opened], 1
	jc @@ok
%endif
		mov	ax,[e_dump_height]
		mov	cl,4
		shl	ax,cl
		add	ax,[dump_offs]
		cmp	si,ax
		jb	@@ok
		mov	ax,si
		sub	ax,[dump_offs]
		and	ax,0fff0h
		add	[dump_offs],ax
		mov	dh,0
@@ok:
		call	shrink_dump
		jmp	draw_screen
@@enlarge:
		mov	es,[video_seg]
		xor	di,di
		cld
		mov	cx,SCR_COLS*SCR_ROWS
		test	byte [cmd_line_options],CMD_EGA
		jz	@@not_ega
		add	cx,SCR_COLS*(MAX_SCR_ROWS-SCR_ROWS)
@@not_ega:
		mov	ax,atMainWindow*100h+20h
	rep	stosw
		call	expand_dump

;=============================================================================
; _update_dump
;-----------------------------------------------------------------------------
_update_dump:
		cmp	word [data_watch_proc],0
		je	update_dump
		jmp	word [data_watch_proc]
update_dump:
%if _TERMINAL
	rol byte [terminal_opened], 1
	jc .ret
%endif
		call	pushr
		mov	si,[dump_offs]
		mov	es,[video_seg]

		mov	ax,si
;-----------------------------------------------------------------------------
; Draw line of hex digits above dump
;-----------------------------------------------------------------------------
		mov	ah,atDumpAddr
		mov	bp,[dump_start]
		mov	di,bp
		sub	bp,SCR_COLS*2-(DUMP_X+13+16*3)*2
		sub	di,SCR_COLS*2-(DUMP_X+12)*2
		mov	cx,16
@@next:
		and	al,0fh
		push	ax
		cmp	al,10
		sbb	al,69h
		das
		stosw
		xchg	di,bp
		stosw
		xchg	di,bp
		pop	ax
		scasw
		scasw
		inc	ax
		loop	@@next
;-----------------------------------------------------------------------------
		mov	ds,[dump_seg]
		mov	di,[cs:dump_start]
		mov	cx,[cs:dump_height]
		sub	dl,16
		neg	dl

		push	dx

@@next_row:
		mov	bh,atDumpAddr
		mov	ax,ds
		call	fill_word
		mov	al,':'
		mov	ah,bh
		stosw
		mov	ax,si
		call	fill_word
		scasw
		scasw
		mov	bl,atDumpValue
		mov	ah,atDumpCursor
		push	cx
		mov	cx,16
@@next_col:
		mov	bh,bl
		lodsb
		or	al,al
		jnz	@@nz
		mov	bh,atDumpZero
@@nz:
		test	byte [cs:edit_mode],ED_DUMP
		jz	@@not_cur_1
		cmp	dh,0
		jne	@@not_cur_1
		cmp	dl,cl
		jne	@@not_cur_1
		mov	bh,ah
@@not_cur_1:
		call	fill_byte
		scasw
		loop	@@next_col
		pop	cx
		dec	dh
		add	di,(SCR_COLS-59)*2
		loop	@@next_row

		pop	dx
		mov	bh,atDumpValue
		mov	bl,atDumpCursor

		mov	si,[cs:dump_offs]
		mov	di,[cs:dump_start]
		add	di,61*2			;!!!
		mov	cx,[cs:dump_height]
@@next_row_1:
		push	cx
		mov	cx,16
@@next_col_1:
		mov	ah,bh
		lodsb
	rol byte [cs:option_cpdump], 1
	jnc @F			; not set, hide CP-dependent -->
	test al, al		; zero ?
	jnz @FFF		; no -->
		; If zero, fall through into the next check
		;  and eventually replace this byte too.
@@:
	cmp al, 32
	jb @F			; if below, replace
	cmp al, 127
	jb @FF			; if above-or-equal, fall through and replace
@@:
		mov	ah,atDumpZero
		mov	al,'.'
@@:
		test	byte [cs:edit_mode],ED_DUMP
		jz	@@not_cur_2
		cmp	dh,0
		jne	@@not_cur_2
		cmp	dl,cl
		jne	@@not_cur_2
		mov	ah,bl
@@not_cur_2:
		stosw
		loop	@@next_col_1

		pop	cx
		dec	dh
		add	di,(SCR_COLS-16)*2
		loop	@@next_row_1

;-----------------------------------------------------------------------------
		test	byte [cs:edit_mode],ED_DUMP
		jz	@@quit
		mov	cx,[cs:dump_seg]
		mov	dx,[cs:dump_address]
		mov	di,DUMP_START-SCR_COLS*4+10
		cmp	byte [cs:large_dump],1
		jne	@@show
		xor	di,di
		call	show_address_2
		jmp	@@quit
@@show:
		call	show_address
@@quit:
;-----------------------------------------------------------------------------

		call	popr
.ret:
		ret
;=============================================================================
; update_stack
;-----------------------------------------------------------------------------
update_stack:
		cmp	byte [large_dump],0
		jz	@@ok
		jmp	@@quit
@@ok:

		call	pushr

		mov	ax,[cpu_regs+REG_SP]
		test	byte [edit_mode],ED_STACK
		jnz	@@0001
		mov	[initial_sp_val],ax
@@0001:

		mov	di,STK_START
		mov	cx,[stack_height]
		mov	si,[initial_sp_val]
		mov	bp,[initial_bp_disp]
		mov	es,[video_seg]
		mov	ds,[cpu_regs+REG_SS]

@@next:
		mov	ax,atStackAddr*256+'S'
		stosw
		stosw
		mov	al,':'
		stosw
		xchg	ax,bx		;Instead of MOV BH,AH
		mov	ax,si
		call	fill_word	;SP value

	mov ah, atStackName
	drawchoice 10h, '>'
	mov al, byte [cs:DRAWADDRESS]
		cmp	si,[cs:cpu_regs+REG_SP]
		je	@@st0
		mov	al,' '
@@st0:
		stosw
		mov	bh,atStackValue
		lodsb
		mov	ah,al
		lodsb
		xchg	ah,al
		call	fill_word	;[SS:SP]

		scasw
		scasw

		mov	ax,atStackAddr*256+'+'
		mov	bx,bp
		or	bp,bp
		jns	@@st1
		mov	al,'-'
		neg	bx
@@st1:
		stosw
		xchg	ax,bx		;BH := AH (color); AX := BX (displ-t)
		call	fill_word	;Displacement
		scasw
		mov	bx,bp
		add	bx,[cs:cpu_regs+REG_BP]
		mov	al,[bx]
		mov	ah,[bx+1]
		mov	bh,atStackValue
		call	fill_word	;[SS:BP+xxxx]
		add	bp,2

		sub	di,(SCR_COLS+24)*2
		loop	@@next

		test	byte [cs:edit_mode],ED_STACK
		jz	@@dont_show
		mov	cx,[cs:cpu_regs+REG_SS]
		mov	bp,[cs:stack_mode]
		mov	dx,word [cs:bp+3]
		mov	di,DUMP_START-SCR_COLS*4+10
		call	show_address
@@dont_show:

		call	popr
@@quit:
		ret
;=============================================================================
; update_regs_color
;-----------------------------------------------------------------------------
update_regs_color:
		cmp	byte [regs_saved],1
		je	@@exit
		inc	byte [regs_saved]
		call	save_regs
		call	update_registers
@@exit:
		ret
;=============================================================================
; display_focus
;-----------------------------------------------------------------------------
display_focus:
		inc	byte [focus]
		jmp	update_commands
;=============================================================================
; hide_focus
;-----------------------------------------------------------------------------
hide_focus:
		dec	byte [focus]
		jmp	update_commands
;=============================================================================
; edit_registers
;-----------------------------------------------------------------------------
edit_registers:
		call	hide_focus
		call	update_regs_color
		call	large_cursor

@@start_edit:
		mov	si,reg16_pos_table
		cmp	byte [regs_mode],1
		jne	@@skip_32
		mov	si,reg32_pos_table
@@skip_32:
		xor	dx,dx

@@next:
		push	dx
		add	dx,[si]
		call	gotoxy
		pop	dx

@@read_key:
		call	read_key

		mov	di,@@actions - 3
		call	action_short_jump
		jc	@@other
		add	di,@@first_action
		jmp	di

@@first_action:
; 0..9, A..F, a..f
@@other:
		call	ascii_to_bin
		jc	@@next
		mov	di,[si+REG_POS_LINK]
		call	set_half_byte
		inc	dl
		cmp	dl,4
		jae	@@move

@@update:
;		call	update_regs_and_cmds
		call	update_screen	;We need to update stack window also...
		jmp	@@next
; Tab
@@move:
		mov	al,[si+REG_POS_RIGHT]
		xor	dx,dx
		jmp	@@get_next
; Shift+Tab
@@move_back:
		mov	al,[si+REG_POS_LEFT]
		mov	dl,0
		jmp	@@get_next
; Right
@@right:
		inc	dl
		cmp	dl,3
		ja	@@move
		jmp	@@next
; Left
@@left:
		or	dl,dl
		jnz	@@dec
		mov	al,[si+REG_POS_LEFT]
		mov	dl,3
		jmp	@@get_next
@@dec:
		dec	dl
		jmp	@@next
; Up
@@up:
		mov	al,[si+REG_POS_UP]
		jmp	@@get_next
; Down
@@down:
		mov	al,[si+REG_POS_DOWN]
		jmp	@@get_next
; Alt+F10
@@local_menu:
		push	si
		mov	si,regs_local_struc
		call	_local_menu
		pop	si
		jc	@@next
		jmp	ax
; Ctrl+R
@@switch_reg_mode:
		call	switch_reg_mode
		jmp	@@start_edit
; Ctrl+A
@@restore_all:
		push	si
		call	restore_all
		pop	si
		jmp	@@next
; Ctrl+T
@@restore_flags:
		call	restore_flags
		jmp	@@next
; Ctrl+C
@@restore_csip:
		call	restore_csip		;[0leg, 2006-12-31]
		jmp	@@next			;Was: restore_flags
; Ctrl+I
@@inc_reg:
		mov	di,[si+REG_POS_LINK]
		cmp	byte [regs_mode],1
		je	@@inc_32
@@inc_16:
		inc	word [di]
		jmp	@@update
@@inc_32:
		cmp	di,cpu_regs+REG_CS
		jae	@@inc_16
		call	align_reg
	add word [di], 1
	adc word [di + 2], 0
		jmp	@@update
; Ctrl+D
@@dec_reg:
		mov	di,[si+REG_POS_LINK]
		cmp	byte [regs_mode],1
		je	@@dec_32
@@dec_16:
		dec	word [di]
		jmp	@@update
@@dec_32:
		cmp	di,cpu_regs+REG_CS
		jae	@@dec_16
		call	align_reg
	sub word [di], 1
	sbb word [di + 2], 0
		jmp	@@update
; Ctrl+Z
@@zero_reg:
		mov	di,[si+REG_POS_LINK]
		call	align_reg
		mov	word [di],0
		mov	word [di+2],0
		jmp	@@update
; Ctrl+O
@@restore_reg:
		mov	di,[si+REG_POS_LINK]
		call	align_reg
		mov	ax,[di+save_cpu_regs-cpu_regs]
		mov	[di],ax
		mov	ax,[di+2+save_cpu_regs-cpu_regs]
		mov	[di+2],ax
		jmp	@@update
; ESC, Enter
@@quit:
		call	no_cursor
		call	update_screen
		jmp	display_focus

@@get_next:	cbw
		shl	ax,1
		shl	ax,1
		shl	ax,1
		add	si,ax
		jmp	@@update

@@actions	dw	kbEsc
		db	@@quit		- @@first_action
		dw	kbF10
		db	@@quit		- @@first_action
		dw	kbAltX
		db	@@quit		- @@first_action
		dw	kbEnter
		db	@@quit		- @@first_action
		dw	kbTab
		db	@@move		- @@first_action
		dw	kbShiftTab
		db	@@move_back	- @@first_action
		dw	kbF12
		db	@@move		- @@first_action
		dw	kbF11
		db	@@move_back	- @@first_action
		dw	kbRight
		db	@@right		- @@first_action
		dw	kbLeft
		db	@@left		- @@first_action
		dw	kbUp
		db	@@up		- @@first_action
		dw	kbDown
		db	@@down		- @@first_action
		dw	kbCtrlA
		db	@@restore_all	- @@first_action
		dw	kbCtrlO
		db	@@restore_reg	- @@first_action
		dw	kbCtrlT
		db	@@restore_flags	- @@first_action
		dw	kbCtrlC
		db	@@restore_csip	- @@first_action
		dw	kbCtrlR
		db	@@switch_reg_mode - @@first_action
		dw	kbCtrlI
		db	@@inc_reg	- @@first_action

		dw	kbCtrlE
		db	@@inc_reg	- @@first_action
	; "Erhoehen". Ctrl+I may be indistinguishable from Tab!

		dw	kbCtrlD
		db	@@dec_reg	- @@first_action
		dw	kbCtrlZ
		db	@@zero_reg	- @@first_action
		dw	kbAltF10
		db	@@local_menu	- @@first_action
		dw	0

_switch_reg_mode	equ	@@switch_reg_mode
_restore_all		equ	@@restore_all
_restore_flags		equ	@@restore_flags
_restore_csip		equ	@@restore_csip
inc_reg			equ	@@inc_reg
dec_reg			equ	@@dec_reg
zero_reg		equ	@@zero_reg
restore_reg		equ	@@restore_reg

;=============================================================================
; set_half_byte
;-----------------------------------------------------------------------------
; Changes nybble in a word.
; In:	DL -- number of nybble: 0..3
;	DS:DI -> word to edit
;	AL -- value of nybble: 0..F
; Out:	DS:DI -> changed word
; Modf:	AX, BX, CL
; Call:	---
; Use:	---
;
set_half_byte:
		mov	bx,0fff0h
		mov	cl,3
		sub	cl,dl
		shl	cl,1
		shl	cl,1
		rol	bx,cl
		and	bx,[di]
		mov	ah,0
		shl	ax,cl
		or	ax,bx
		mov	[di],ax
		ret
;=============================================================================
; toggle_flag
;-----------------------------------------------------------------------------
toggle_flag:
		call	get_flag_mask
		xor	[cpu_regs+REG_FL],ax
		ret

get_flag_mask:
		mov	cl,[reg_shift+bx]
		mov	ax,1
		shl	ax,cl
		ret
;=============================================================================
; edit_flags
;-----------------------------------------------------------------------------
edit_flags:
		call	hide_focus
		call	update_regs_color
		call	large_cursor
		xor	bx,bx
@@next:
		mov	dx,(STK_BOT_Y-1)*256+FLAG16_X+2
		sub	dh,[stack_height]
		add	dl,bl
		add	dl,bl
		add	dl,bl
		call	gotoxy
		call	read_key

		mov	di,@@actions - 3
		call	action_short_jump
		jc	@@next
		add	di,@@first_action
		jmp	di

@@first_action:
; 0
@@clear:
		call	get_flag_mask
		not	ax
		and	[cpu_regs+REG_FL],ax
		jmp	@@update
; 1
@@set:
		call	get_flag_mask
		or	[cpu_regs+REG_FL],ax
		jmp	@@update
; Alt+F10
@@local_menu:
		mov	si,flags_local_struc
		call	_local_menu
		jc	@@next
		call	ax
		jmp	@@update
; Space
@@toggle:
		call	toggle_flag
@@update:
		call	update_regs_and_cmds
		jmp	@@next
; Right
@@right:
		inc	bl
		cmp	bl,NUM_FLAGS
		jb	@@next
		mov	bl,0
		jmp	@@next
; Left
@@left:
		dec	bl
		jns	@@next
		mov	bl,NUM_FLAGS-1
		jmp	@@next
; Ctrl+T
@@restore_flags:
		call	restore_flags
		jmp	@@next
; ESC, Enter
@@quit:
		call	no_cursor
		jmp	display_focus


@@actions	dw	kbEsc
		db	@@quit		- @@first_action
		dw	kbF10
		db	@@quit		- @@first_action
		dw	kbAltX
		db	@@quit		- @@first_action
		dw	kbEnter
		db	@@quit		- @@first_action
		dw	kbRight
		db	@@right		- @@first_action
		dw	kbLeft
		db	@@left		- @@first_action
		dw	kbDown
		db	@@clear		- @@first_action
		dw	kbUp
		db	@@set		- @@first_action
		dw	0b00h+'0'
		db	@@clear		- @@first_action
		dw	0200h+'1'
		db	@@set		- @@first_action
		dw	3900h+' '
		db	@@toggle	- @@first_action
		dw	kbCtrlT
		db	@@restore_flags	- @@first_action
		dw	kbAltF10
		db	@@local_menu	- @@first_action
		dw	0

%if _TERMINAL
connect_terminal:
	rol byte [terminal_connected], 1
	jc .already

	mov si, serialdata
	mov ax, [config_serial_radio_selected]
	cmp ax, 4
	je .manual

	mov si, ax
	add si, si			; times 2
	mov ax, si
	add si, si			; times 4
	add si, ax			; times 6
	add si, fixed_serialdata	; -> array entry

.manual:
	mov di, actual_serialdata
	movsw
	movsw
	movsw				; move over manual or fixed data

	call init_terminal_size

	push es
	mov ax, 40h
	mov es, ax
        mov bx,[es:84h]
        xor bh,bh
        cmp bl,24		; we need at least 25 rows on screen
        jnb .validrows
        mov bl,24		; if we have less, just pretend to have 25 !
.validrows:
	mov [rows], bx
        mov ah,bl
	mov bx, SCR_COLS
        mov al,bl
        dec ax
	mov [rendxy], ax	; maximum values of X and Y for ANSI
        mov ax,[cs:rows]
        mul bl
        add ax,ax
	mov [rend], ax		; store pointer to end of screen
        mov bx,[es:63h]
	mov [crtport], bx	; store CRT port base - needed for Vretrace

	mov al,[cs:rows]
	sub al,[cs:rfirstrow]
	mov [rscrollreg+1],al	; save max relative line number for scrolling
	mov [asavescrollreg],al

	push es
	 push cs
	 pop es
	mov ax, [attr]
	mov di, buffers
	mov cx, buffers_size / 2
	rep stosw
	pop es

		call SPEEDSET	; set up serial communications speed
		call IRQ_INST   ; ***** set up interrupt driven serial
	mov byte [modlines+1],0bh	; turn off: res res ??? loopback
				; (leave out2) out1 (leave rts) (leave dtr)
        or word [modlines],3  ; turn RTS/DTR on: we are now ready
        or word [modlines],8  ; turn RTS/DTR on: we are now ready
                call MOD_LINES	; inform UART about this

	pop es

	mov byte [terminal_connected], -1
	and word [options_terminal_sub.disconnect], 0
	mov word [options_terminal_sub.connect], DISABLED

.already:
	retn


disconnect_terminal:
	rol byte [terminal_connected], 1
	jnc @F
	mov byte [modlines+1],0fch	; leave everything on, but...
	and word [modlines],0fcfch	; turn RTS/DTR off
	call MOD_LINES
	call IRQ_UNINST
	mov byte [terminal_connected], 0
	and word [options_terminal_sub.connect], 0
	mov word [options_terminal_sub.disconnect], DISABLED
@@:
	retn


open_terminal:
	call init_terminal_size

	rol byte [option_terminal_autoconnect], 1
	jnc @F
	call connect_terminal
@@:

	push es
		mov	es,[video_seg]
	; needed for initgeo, SHOW_TX (called in SENDME), SHOW_STATS, SHOW_PROTO

                call initgeo	; see below: find buffer size, clear screen
		call hide_focus
		mov byte [terminal_opened], -1
		mov byte [terminal_active], -1
		call update_terminal
		call SHOW_PROTO
		xor dx, dx

@@next:
	rol byte [option_terminal_largecursor], 1
	jc @F
	call small_cursor
	jmp @FF
@@:
	call large_cursor
@@:
	call setcursor

        PUSHf
        cli
        mov dx,[cs:comdta]   ; ah=3 dx=comport int 14h
        add dl,5
        in al,dx
        mov ah,al
        inc dx
        in al,dx        ; Status AH/AL
        mov [comstats],ax	; (( ah test 1 nz if data waiting ))
        POPf

.SHOW_DATA_RECEIVED:
        test word [hshake],2          ; hshake test 2: send RTS
        jz .no_hs1
	mov byte [modlines+1],(0ffh | RTSSEND)	; clear used flags
        mov byte [modlines],8	; set no flags	; else send no RTS
				; ok, always set one flag: 8 is OUT2
				; and is used as int enable...!
                call MOD_LINES	; inform modem about this
.no_hs1:         call GETSH	; get keyboard shift status
	jc .raus

.shownix:
	mov ax,[cs:comstats]
	cmp ax,[cs:comstats2]
	jz .nonewstats		; only show comstats if changed
	mov [comstats2],ax
                call SHOW_STATS ; Modem-/Line-Status comstats
.nonewstats:
                call GETSH	; get keyboard shift status
	jc .raus

.GET_KEY_TO_TRANSMIT:
        mov BP,[cs:hshake]		; create local copy of hshake !!!
	and BP,0ff3fh			; clear BREAK and FrameERROR
        mov ah,1        ; check for key, peek what key it is
                int 16h ; Z if none waiting
        jnz .n_sendnix
.y_sendnix:
	jmp .sendnix	; nothing to send
.n_sendnix:

		call read_key
		mov	di,@@actions - 4
		call	action_long_jump
		jc	@@not_found
		add	di,@@first_action
		jmp	di

@@not_found:
        mov BP,[cs:hshake]		; create local copy of hshake !!!
	and BP,0ff3fh			; clear BREAK and FrameERROR
	test al, al
        jz .y_sendnix	; only send printable chars for now
			; <<< *** improve to send cursor thing as well
			; <<< *** (keypad things are sent as ESC O x in alt
			; <<< *** keypad mode, cursor keys are sent as
			; <<< *** ESC A/B/C/D for UP/DOWN/R/L in normal
			; <<< *** and as ESC 0 A/B/C/D in application
			; <<< *** cursor mode. So this would send 1..3 chars
			; <<< *** for each cursor/keypad event !
                call SENDME	; send char from special key handler
		jmp @@next

@@first_action:

; Alt+Z
@@zoom_dump:
		call zoom_dump
		call update_terminal

@@jump_to_next:
		jmp	@@next


@@enter:
.KEY_IS_ENTER:
        mov bx,[cs:crpolicy]
        shl bx,1	; get a key SEQUENCE to send ENTER as 1 or 2 chars!
        mov ax,[cs:send_cr+bx]
	test ah, ah	; if we have a "sequence" of only 1 char, treat it
			; as if it were a normal key
        jz .SEND_NORMAL_KEY
			; (index to list selected by the crpolicy)
        PUSH ax
                call SENDME	; send char in ENTER sequence to remote side
        POP ax
        mov al,ah

.SEND_NORMAL_KEY:
                call GETSH	; get keyboard shift status
	jc .raus
                call SENDME     ; send that key to the remote side!
	cmp ax,-1
	jz .KEY_WAS_LOST
	jmp short .sendok

.KEY_WAS_LOST:
	mov BP,[cs:hshake]	; create a local copy !!!
        and BP, 0ff2fh          ; clear the BREAK and FrameERROR status bits
				; and the JAMMED status bit (now 00x0xxxx)
        or BP,10h		; store "we are jammed" status
                call GETSH	; get keyboard shift status
	jc .raus ; <<<
        jmp short .senddone      ; keys will pile up in keyboard buffer
				; until the traffic jam is over
				; (problem: menu keys will do the same!)
	; key was successfully sent:
.sendok:
	mov BP,[cs:hshake]	; create a local copy !!!
        and BP, 0ff2fh          ; clear the BREAK and FrameERROR status bits
				; and the JAMMED status bit (now 00x0xxxx)
	jmp .senddone

	; there was not even a key to be sent:
.sendnix:
	pushf
	call IRQ_READCHECK	; data available for read ?
	jne .dontidle		; yes, do not idle -->
	call idle
.dontidle:
	call terminal_receive_to_buffer
	popf
.senddone:

.BREAK_HANDLING:
	; hshake: bits are SSUSxxyU (S is status, U is from user,
	;         x is setting of rx logic from user, y is setting
	;         of tx flag send policy from user)
        mov ax,[cs:comstats]
        test ax,800h    ; Error/Break frame?
        jz .nobrk1
        or BP,80h
.nobrk1: test ax,200h    ; Break (over) ?
        jz .nobrk2
        or BP,40h
.nobrk2: cmp BP,[cs:hshake]	; if hshake has changed, update display!!!
        jz .nixpro
        mov [hshake],BP
                call SHOW_PROTO	; show current comm/crlf/handshake setup
.nixpro: test word [hshake],2		; if hshake test 2, turn on RTS
        jz .no_hs2
	mov byte [modlines+1],0ffh	; and: leave all as-is
        or byte [modlines],RTSSEND	; or: RTS on
                call MOD_LINES		; tell UART about this
.no_hs2:


; <<<< *** not very useful yet: %define PREVENT_JAMMED_KEYBOARD 1
%ifdef PREVENT_JAMMED_KEYBOARD
 %assign DEF_PREVENT_JAMMED_KEYBOARD_ALWAYS 1
%else
 %assign DEF_PREVENT_JAMMED_KEYBOARD_ALWAYS 0
%endif
%assign _PREVENT_JAMMED_KEYBOARD_ON_STAU 1

%if _PREVENT_JAMMED_KEYBOARD_ON_STAU || DEF_PREVENT_JAMMED_KEYBOARD_ALWAYS
 %if _PREVENT_JAMMED_KEYBOARD_ON_STAU && (DEF_PREVENT_JAMMED_KEYBOARD_ALWAYS == 0)
	test byte [hshake], 10h
	jz .prevent_not_jammed
 %endif
	push es
	push bx
	push cx
	mov ax,40h
	mov es,ax
	xor cx,cx
	mov ax,[es:1ah]	; keyb: pointer to next char to read, read pointer
	mov bx,[es:1ch]	; keyb: pointer to first free slot, write pointer
.prjam0:	cmp ax,bx	; if both are the same, buffer is empty
	jz .prjam1
	inc cx	; one more key waiting to be read
	add ax,2	; as if we were reading...
	cmp ax,[es:82h]
	jb .prjam0	; if not below, we reached the end of buffer...
	mov ax,[es:80h]	; ...and wrap around to the beginning
	jmp short .prjam0
.prjam1:	mov ax,[es:82h]
	sub ax,[es:80h]
	shr ax,1	; now ax is size of buffer, in chars
	shr ax,1	; now it is 50% of that
	cmp cx,ax
	jb .prjam2	; less than 50% in use, good
	mov ah,0
		int 16h	; read one key from buffer, then discard it!
			; could also use func 04h to flush all the keys...
.prjam2:
	pop cx
	pop bx
	pop es
.prevent_not_jammed:
%endif
; <<<< *** end of PREVENT_JAMMED_KEYBOARD

	jmp @@next

; ESC
.raus:
@@quit:
		mov byte [terminal_active], 0
		mov byte [flag_exit], 0
		call	no_cursor
		call	shrink_dump
		call	draw_screen
		call	display_focus
		pop	es
		ret

@@right:
	mov al, 'C'
	jmp .right

@@left:
	mov al, 'D'
	jmp .left

@@up:
	mov al, 'A'
	jmp .up

@@down:
	mov al, 'B'
	jmp .down

@@page_up:
	mov al, 35h
	jmp .pageup

@@page_down:
	mov al, 36h
	jmp .pagedown

@@insert:
	mov al, 32h
	jmp .insert

@@del:
	mov al, 33h
	jmp .del

; @@backspace:
;	mov al, 8

@@home:
	mov al, 31h
	jmp .home

@@end:
	mov al, 34h
	jmp .end

.right:
.down:
.left:
.up:
.arrow:
	push ax
	mov al, 1Bh
	call SENDME
	mov al, '['
	call SENDME
	pop ax
.send_then_senddone:
	call SENDME
	jmp .senddone

.insert:
.del:
.home:
.end:
.pageup:
.pagedown:
	push ax
	mov al, 1Bh
	call SENDME
	mov al, '['
	call SENDME
	pop ax
	call SENDME
	mov al, 7Eh
	jmp .send_then_senddone


@@f1:	call SHOW_ROTATE	; F1 circulates the status disp lines
        jmp @@fnone		; Disp (returns nothing)

@@f2:    inc word [linespeed]	; F2 increases the line speed one step
        and word [linespeed],7  ; Sp+
	push word [linespeed]
	pop word [baudlistchoice]
                call SPEEDSET
        jmp @@fstat

@@f3:    xor word [hshake],20h	; F3 toggles sending of BREAK
                call TX_BRK
        jmp @@fstat              ; BRK

@@f4:    mov al,3                ; ETX:  F4 sends an ETX (^C) character
        jmp short @@fret		; RETURNS a char

@@f5:    mov ax,[cs:crpolicy]	; tCR: F5 changes the way ENTER is sent
	inc ax			; (4 ways, F5 circulates through them)
	test ax,3
	jnz @@f5x	; store new value
	sub ax,4	; we overflowed our 2 bits, fixing...
@@f5x:   mov [crpolicy],ax	; store new value
        jmp short @@fstat

@@f6:    xor word [crpolicy],8000h	; F6 changes the way CR and LF are
        jmp short @@fstat        ; rCR	displayed: as is or in a way that
					; either of them acts like CR+LF

@@f7:    xor word [hshake],8	; CTS: F7 toggles waiting for CTS handshake
	test word [hshake],8	; + DSR
	jnz @@fstat
	xor word [hshake],4	; F7 now toggles waiting for DSR handshake
        jmp short @@fstat	; after toggling the CTS handshake to off
				; (so we do a 4-cycle here...)

; @@f8:    xor word [hshake],4	; DSR: F8 toggles waiting for DSR handshake
;         jmp short @@fstat

@@f8:	mov word [flag_exit],1	; EXIT: F8 to exit program
	jmp short @@fstat

@@f9:    xor word [hshake],2	; RTS: F9 toggles sending RTS handshake
        jmp short @@fstat	; (see lines with "RTSSEND" in terminal.asm)

@@f10:   mov ax,[cs:hshake]	; DTR: F10 toggles sending DTR handshake
        xor ax,1
        mov [hshake],ax
        mov word [modlines],0fe08h	; mask DTR to 0, and set it...
				; (8 means always set OUT2, int enable)
        and ax,1
        or [modlines],ax		; ...to value from AX then
                call MOD_LINES	; changes are sent to the modem right away

@@fstat:         call SHOW_PROTO	; protocol setup has changed, update display
@@fnone: xor ax,ax		; return no char
@@fret:
	jmp @@next


	align 4, db 0
@@actions	dw	kbEsc
		dw	@@quit		- @@first_action
		dw	kbAltX
		dw	@@quit		- @@first_action
		dw	kbF10
		dw	@@quit		- @@first_action

		dw	kbF1
		dw	@@f1		- @@first_action
		dw	kbF2
		dw	@@f2		- @@first_action
		dw	kbF3
		dw	@@f3		- @@first_action
		dw	kbF4
		dw	@@f4		- @@first_action
		dw	kbF5
		dw	@@f5		- @@first_action
		dw	kbF6
		dw	@@f6		- @@first_action
		dw	kbF7
		dw	@@f7		- @@first_action
		dw	kbF8
		dw	@@f8		- @@first_action
		dw	kbF9
		dw	@@f9		- @@first_action
		dw	kbF11
		dw	@@f10		- @@first_action

		dw	kbEnter
		dw	@@enter		- @@first_action
		dw	kbRight
		dw	@@right		- @@first_action
		dw	kbLeft
		dw	@@left		- @@first_action
		dw	kbUp
		dw	@@up		- @@first_action
		dw	kbDown
		dw	@@down		- @@first_action
		dw	kbPgUp
		dw	@@page_up	- @@first_action
		dw	kbPgDn
		dw	@@page_down	- @@first_action
		dw	kbIns
		dw	@@insert	- @@first_action
		dw	kbDel
		dw	@@del		- @@first_action
		; dw	kbBackSpace
		; dw	@@backspace	- @@first_action
		dw	kbHome
		dw	@@home		- @@first_action
		dw	kbEnd
		dw	@@end		- @@first_action
		; dw	kbAltF10
		; dw	@@local_menu	- @@first_action
		dw	kbAltZ
		dw	@@zoom_dump	- @@first_action

		dw	kbCtrlV
		dw	@@scroll	- @@first_action

		dw	0			; end of table


@@scroll:
scroll_terminal:
	call update_terminal
	call no_cursor
	call pushr

		; First, find the last terminal_linebreaks.amount
		;  linebreaks by parsing the whole terminal buffer.
		; Linebreaks occur when an LF is encountered or a
		;  codepoint causes the line to overflow into the
		;  next line, which we detect in SHOW_RX / TTY.
	mov di, word [terminal_buffer.end]

		; This entrypoint parses the terminal buffer from
		;  the start until reaching the offset passed in di.
.loop_until_di:
	mov si, word [terminal_buffer.start]

		; Reset position to start position.
	mov ax, word [rstartxy]
	mov word [rposxy], ax
	mov word [asaveposxy], ax
	mov ax, word [rstart]
	mov word [rpos], ax

		; Indicate we are counting lines, so no actual output
		;  is done by the TTY function. (Not that useful but
		;  might speed up things a bit.)
	mov byte [terminal_counting_lines], -1

		; Reset the terminal_linebreaks circular buffer.
	mov bx, terminal_linebreaks
	mov word [terminal_linebreaks.first], bx
	mov word [terminal_linebreaks.last], bx

		; The very first boundary in the terminal buffer should
		;  count as a linebreak. So jump into the middle of the
		;  first loop here, to record this offset as a linebreak.
	jmp .loop_got_linebreak
@@:
	cmp si, di				; end of buffer space to scan ?
	je @F					; yes -->
	mov es, word [terminal_buffer]		; prepare to load from buffer
	es lodsb				; load from buffer
	mov es, word [video_seg]		; (make sure correct segment)
	mov byte [terminal_linebreak_detected], 0
						; clear flag
	call SHOW_RX				; process codepoint
	rol byte [terminal_linebreak_detected], 1
						; flag is set now ?
	jnc @B					; no, just loop -->

.loop_got_linebreak:
	mov word [bx], si			; store offset in free slot
	call .next_linebreak			; point to next slot
	cmp bx, word [terminal_linebreaks.first]; overflow ?
	jne @B					; no, just loop next -->
	xchg bx, word [terminal_linebreaks.first]
	call .next_linebreak			; point to next slot
	xchg bx, word [terminal_linebreaks.first]
						; update .first offset
	jmp @B					; and loop -->

@@:
	mov word [terminal_linebreaks.last], bx	; store end

	mov cx, word [dump_height]		; get height of terminal window
						;  (not including status line)
@@:
	cmp bx, word [terminal_linebreaks.first]; are we at the first linebreak ?
	je @F					; yes, exit loop -->
	call .prev_linebreak			; go back
	loop @B					; loop up to height times
@@:

.loop:
		; Reset position.
	mov ax, word [rstartxy]
	mov word [rposxy], ax
	mov word [asaveposxy], ax
	mov ax, word [rstart]
	mov word [rpos], ax
	mov byte [terminal_counting_lines], 0

		; Clear the terminal display to avoid remains.
	call pushr
	call update_terminal.clear
	call popr

		; Start after selected linebreak.
	mov si, word [bx]
	jmp @F					; skip .down handling
.down:
	mov es, word [video_seg]
	mov al, 13
	call SHOW_RX
	mov al, 10
	call SHOW_RX

	mov byte [terminal_next], 0
	mov dx, bx
	nop		; mysterious NASM bug needs this
	cmp bx, word [terminal_linebreaks.last]
	je .down_2
	call .next_linebreak
	cmp bx, word [terminal_linebreaks.last]
	je .down_2
	jmp @F
.down_2:
	mov byte [terminal_next], -1
	mov bx, dx

		; This loop actually displays data from the terminal.
@@:
	mov di, word [terminal_buffer.end]	; limit output to buffer end
	mov byte [terminal_do_not_scroll], -1	; do not actually scroll
	mov byte [terminal_scroll_detected], 0	; reset flag
	jmp .loop_middle			; if si == last linebreak,
						;  handle that condition first
@@:
	cmp si, di				; end of buffer space ?
	je @F					; yes, end loop now -->
	mov es, word [terminal_buffer]		; prepare loading
	es lodsb				; load
	mov es, word [video_seg]		; set segment to display to
	call SHOW_RX				; actually display
.loop_middle:
	push bx
	mov bx, word [terminal_linebreaks.last]	; -> behind last entry
						;  (there will always be at
						;  least one entry here)
	call .prev_linebreak			; -> last entry
	cmp si, word [bx]			; matches current offset ?
	pop bx
	je .loop_2				; yes, do special loop -->
	rol byte [terminal_scroll_detected], 1	; end of display reached ?
	jnc @B					; not yet, loop -->

@@:
	rol byte [terminal_next], 1		; need to advance .last now ?
	jnc .read_key				; no, go to read key -->
%if 0
	int3					; debugging help
	test word [terminal_buffer], 0
	test word [terminal_linebreaks], 0
	test word [terminal_linebreaks.first], 0
	test word [terminal_linebreaks.last], 0
%endif
						; could show an error here
	jmp .read_key				; go to read key -->

.loop_3:
	rol byte [terminal_next], 1		; do we need to advance bx ?
	jnc @B					; no, just go to read key -->
	mov byte [terminal_counting_lines], -1	; set flag

	push word [rposxy]
	push word [asaveposxy]
	push word [rpos]			; save position

	mov byte [terminal_do_not_scroll], 0	; actually scroll now
	mov es, word [video_seg]
	mov al, 13
	call SHOW_RX
	mov al, 10
	call SHOW_RX				; do scroll one line

	mov byte [terminal_scroll_detected], 0	; reset flag
	mov byte [terminal_do_not_scroll], -1	; don't scroll any longer

	xchg dx, bx				; dx = last bx
	mov bx, word [terminal_linebreaks.last]	; bx -> after last used slot
@@:
	cmp si, di				; end of buffer ?
	je @FF					; yes, abort now -->
	mov es, word [terminal_buffer]
	es lodsb
	mov es, word [video_seg]
	mov byte [terminal_linebreak_detected], 0
	call SHOW_RX
	rol byte [terminal_linebreak_detected], 1
	jnc @F
	mov word [bx], si			; store linebreak offset
	call .next_linebreak			; -> next slot
	mov word [terminal_linebreaks.last], bx	; store pointer
	cmp bx, word [terminal_linebreaks.first]; equal to first offset ?
	jne .loop_3_not_overflow		; no -->
	xchg bx, word [terminal_linebreaks.first]
	call .next_linebreak			; -> next slot
	xchg bx, word [terminal_linebreaks.first]
						; update first offset variable
.loop_3_not_overflow:
@@:
	rol byte [terminal_scroll_detected], 1	; found scrolling ?
	jnc @BB					; not yet, loop -->
@@:
	mov byte [terminal_do_not_scroll], 0	; reset
	mov byte [terminal_counting_lines], 0	; reset
	mov byte [terminal_scroll_detected], -1	; make sure to pass this
	pop word [rpos]
	pop word [asaveposxy]
	pop word [rposxy]			; restore position
	jmp .loop_common


.loop_2:
	rol byte [terminal_scroll_detected], 1	; scroll detected immediately
						;  at matched offset ?
	jc .loop_3				; yes, super special -->
	xchg dx, bx				; dx = last bx
	mov bx, word [terminal_linebreaks.last]	; -> behind last used slot
@@:
	cmp si, di				; at end of buffer ?
	je @FF					; yes -->
	mov es, word [terminal_buffer]
	es lodsb				; load
	mov es, word [video_seg]
	mov byte [terminal_linebreak_detected], 0
	call SHOW_RX				; display
	rol byte [terminal_linebreak_detected], 1
	jnc @F					; no linebreak -->
	mov word [bx], si			; store in next slot
	call .next_linebreak			; -> behind next slot
	mov word [terminal_linebreaks.last], bx	; update offset
	cmp bx, word [terminal_linebreaks.first]; overflow ?
	jne .loop_2_not_overflow		; no -->
	xchg bx, word [terminal_linebreaks.first]
	cmp dx, bx				; last bx at same as first ?
	jne .loop_2_dx_ne_bx			; no -->
	xchg dx, bx
	call .next_linebreak			; step last bx
	xchg dx, bx
.loop_2_dx_ne_bx:
	call .next_linebreak			; step first
	xchg bx, word [terminal_linebreaks.first]
						; save it away
.loop_2_not_overflow:
@@:
	rol byte [terminal_scroll_detected], 1	; scroll detected yet ?
	jnc @BB					; no, loop -->
@@:
.loop_common:
	mov word [terminal_linebreaks.last], bx	; store new behind last
	mov bx, dx				; restore old bx
	rol byte [terminal_next], 1		; special handling ?
	jnc @F					; no -->
	mov byte [terminal_next], 0
	; mov dx, bx
	nop		; mysterious NASM bug needs this
	cmp bx, word [terminal_linebreaks.last]	; at last already ?
	je .down_3				; yes, cannot move up -->
	call .next_linebreak			; -> next slot
	cmp bx, word [terminal_linebreaks.last]	; pointing behind last slot ?
	je .down_3				; yes, reset to point *at* last
	jmp @F					; updated bx successfully
.down_3:
	mov bx, dx				; reset to old (at last slot)
@@:

		; Note that si -> after last displayed codepoint here.
.read_key:
	mov byte [terminal_do_not_scroll], 0	; reset
	call read_key				; get a key
		; This call can store new data in the terminal buffer.
		;  We do not handle this in any way yet.

	cmp ax, kbUp
	je .up
	rol byte [terminal_scroll_detected], 1	; did we end by scrolling ?
	jnc @F					; no, means we are at the end
						;  of the terminal buffer -->
	cmp ax, kbDown
	je .down				; handle scrolling down -->
@@:
	cmp ax, kbCtrlC
	je @F
	cmp ax, kbAltX
	je @F
	cmp ax, kbEsc
	je @F
	cmp ax, kbF10
	jne .read_key				; unknown input, loop -->
@@:
	call popr
	call update_terminal			; restore last output
	jmp open_terminal@@next

.up:
	cmp bx, word [terminal_linebreaks.first]; already at first stored ?
	je @F					; yes, handle specifically -->
	call .prev_linebreak			; just go back one slot
	jmp .loop				;  and select that as start

		; Now we cannot just go back because we were at
		;  the very beginning of the list of linebreaks
		;  in the circular buffer. So instead, find the
		;  linebreak corresponding to the end of what we
		;  last displayed. Then, back up two slots.
		; Magic alert: It is unclear why we have to back
		;  up twice after finding the buffer entry that
		;  contains our offset. However, it does the right
		;  thing this way whereas backing up just once
		;  would result in displaying the same data again.
@@:
	mov bx, word [terminal_linebreaks.last]	; bx -> after last 
@@:
	cmp bx, word [terminal_linebreaks.first]; equal to first ?
	je @F					; yes, cannot go further -->
	call .prev_linebreak			; -> previous slot
	cmp word [bx], si			; is this equal to the end of
						;  what we displayed ?
	jne @B					; not yet, loop -->

		; word [terminal_buffer]:word [bx] -> behind last line displayed
	cmp bx, word [terminal_linebreaks.first]; equal to first ?
	je @F
	call .prev_linebreak			; no, go back another time
		; word [terminal_buffer]:word [bx] -> at last line displayed

@@:
	cmp bx, word [terminal_linebreaks.first]; equal to first ?
	je @F
	call .prev_linebreak			; no, go back another time
		; word [terminal_buffer]:word [bx] -> at second to last line displayed
@@:
	mov di, word [bx]			; -> end of what to display
	jmp .loop_until_di			; now re-detect linebreaks
						;  until this buffer offset

.next_linebreak:
	inc bx
	inc bx					; -> next slot if no wrap
	cmp bx, terminal_linebreaks.end		; after end ?
	jne @F					; not yet -->
	mov bx, terminal_linebreaks		; cycle back to start
@@:
	retn

.prev_linebreak:
	cmp bx, terminal_linebreaks		; at start ?
	jne @F					; no -->
	mov bx, terminal_linebreaks.end		; cycle back to end
@@:
	dec bx
	dec bx					; -> previous slot
	retn
%endif


;=============================================================================
; edit_dump
;-----------------------------------------------------------------------------
edit_dump:
%if _TERMINAL
	xor ax, ax
	xchg al, byte [terminal_opened]
	test al, al
	jz @F
	call draw_screen
@@:
%endif
		or	byte [edit_mode],ED_DUMP
		call	hide_focus

		push	es

@@next_r:
		mov	si,[dump_offs]
		mov	[cs:dump_address],si	;2008-03-23
		mov	es,[dump_seg]
		xor	bx,bx
		xor	dx,dx
		xor	bp,bp

		call	update_dump

@@next:
		call	large_cursor
		push	dx
		or	bp,bp
		jz	@@hex_cur
		add	dl,DUMP_X+11+48+2
		jmp	@@set_cur
@@hex_cur:
		mov	al,dl
		add	dl,al
		add	dl,al
		add	dl,bl
		add	dl,DUMP_X+11
@@set_cur:
		add	dh,[dump_y]
		call	gotoxy
		pop	dx

@@read_key:
		call	read_key
		mov	di,@@actions - 4	;2008-11-03
		call	action_long_jump
		jc	@@not_found
		add	di,@@first_action
		mov	ax,dump_csip
		jmp	di
; 0..9, A..F, a..f
@@not_found:
		or	bp,bp
		jnz	@@fill_char
		call	ascii_to_bin
		jc	@@next
		mov	cl,bl
		xor	cl,1
		shl	cl,1
		shl	cl,1
		mov	ah,0f0h
		rol	ah,cl
		and	ah,[es:si]
		shl	al,cl
		or	ah,al
		mov	[es:si],ah
		cmp	[es:si],ah
		je	@@right
@@not_ram:
		push	si
		mov	si,dump_not_in_ram
		call	_error_message
		pop	si
		jmp	@@update

@@first_action:
; Ctrl+F
@@dump_follow:
		push	si
		mov	si,follow_menu_struc
		jmp	@@exec_menu
; Ctrl+B
@@block:
		push	si
		mov	si,block_menu_struc
		jmp	@@exec_menu
; Alt+F10
@@local_menu:
		push	si
		mov	si,dump_menu_struc

@@exec_menu:
		call	_local_menu
		pop	si
		jc	@@jump_to_next

@@call_action:
		call	pushr
		call	ax
		call	popr
		jmp	@@select_next_jump
; Ctrl+D
@@dump_go_to:
		call	dump_go_to

@@select_next_jump:
		jc	@@jump_to_next

@@jump_to_next_r:
		jmp	@@next_r
; Any char
@@fill_char:
		cmp	al,0
		je	@@jump_to_next
		mov	[es:si],al
		cmp	[es:si],al
		jne	@@not_ram
; Right
@@right:
		or	bp,bp
		jnz	@@cm_1
		xor	bl,1
		jnz	@@update		;jmp @@next ???
@@cm_1:
		inc	si
		inc	dl
		cmp	dl,15
		ja	@@next_line
@@update:
		mov	[cs:dump_address],si	;2008-03-23
		call	update_dump
		call	update_commands
		call	update_stack
		jmp	@@next
; Alt+=
@@dump_reference:
		mov	ax,dump_reference
		jmp	@@call_action
; Alt+L
@@dump_cur_line:
		mov	ax,dump_cur_line
		jmp	@@call_action
; Alt+U
@@dump_user_scr:
		mov	ax,dump_user_screen
		jmp	@@call_action
; Alt+E
@@dump_esdi:
		sub	ax,update_dump_window-dump_csip
; Alt+D
@@dump_dssi:
		sub	ax,update_dump_window-dump_csip
; Alt+B
@@dump_esbx:
		sub	ax,update_dump_window-dump_csip
; Alt+S
@@dump_ssbp:
		sub	ax,update_dump_window-dump_csip
; Alt+A
@@dump_dxax:
		sub	ax,update_dump_window-dump_csip
; Alt+F
@@dump_dsdx:
		sub	ax,update_dump_window-dump_csip
; Alt+I
@@dump_essi:
		sub	ax,update_dump_window-dump_csip
; Alt+J
@@dump_dsdi:
		sub	ax,update_dump_window-dump_csip
; Alt+C
@@dump_csip:
		jmp	@@call_action
; Ctrl+S
@@search:
		mov	ax,data_search
		jmp	@@call_action
; Ctrl+L
@@search_next:
		mov	ax,data_search_next
		jmp	@@call_action
; Alt+Z
@@zoom_dump:
		call	zoom_dump

@@jump_to_next:
		jmp	@@next

@@next_line:
		mov	dl,0
@@down_1:
		inc	dh
		cmp	dh,[dump_height]
		jb	@@update
		add	word [dump_offs],16
		mov	dh,[dump_height]
		dec	dh
		jmp	@@update
; Down
@@down:
		add	si,16
		jmp	@@down_1
; Left
@@left:
		or	bp,bp
		jnz	@@cm_2
		xor	bl,1
		jz	@@update		;jmp @@next ???
@@cm_2:
		dec	si
		dec	dl			;dl always >=0 before dec
		jns	@@update
		mov	dl,15
@@up_1:
		dec	dh
		jns	@@update		;dh always >=0 before dec
		mov	dh,0
		sub	word [dump_offs],16
		jmp	@@update
; Up
@@up:
		sub	si,16
		jmp	@@up_1
; Home
@@home:
		mov	al,dl
		cbw
		mov	dl,0
		jmp	@@upd0
; End
@@end:
		mov	al,dl
		cbw
		add	si,15
		mov	dl,15
@@upd0:
		sub	si,ax
		xor	bl,bl
		jmp	@@update
; Ctrl+Right
@@byte_right:
		inc	word [dump_offs]
		inc	si
		jmp	@@update
; Ctrl+Left
@@byte_left:
		dec	word [dump_offs]
		dec	si
		jmp	@@update
; Tab
@@switch:
		xor	bp,1
		jnz	@@jump_to_next
		xor	bx,bx
		jmp	@@jump_to_next
; Page Up
@@page_up:
		mov	al,-16
		jmp	@@up_dn
; Page Down
@@page_down:
		mov	al,16
@@up_dn:
		imul	byte [dump_height]
		add	[dump_offs],ax
		add	si,ax
		jmp	@@update
; ESC, Enter
@@quit:
		call	no_cursor
		and	byte [edit_mode],~ED_DUMP
		call	shrink_dump
		call	draw_screen
		call	display_focus
		pop	es
		ret
; Alt+F5
@@user_scr:
		call	restore_screen
		call	read_key
		call	save_screen
		jmp	@@next

@@follow_data_seg:
		mov	ax,[es:si]
		mov	[dump_seg],ax
		xor	ax,ax
		jmp	@@follow_data

@@follow_far_data:
		mov	ax,[es:si+2]
		mov	[dump_seg],ax

@@follow_near_data:
		mov	ax,[es:si]
@@follow_data:
		mov	[dump_offs],ax
		call	update_dump
		clc
		ret

@@follow_code_seg:
		mov	ax,[es:si]
		mov	[unasm_seg],ax
		xor	ax,ax
		jmp	@@follow_code

@@follow_far_code:
		mov	ax,[es:si+2]
		mov	[unasm_seg],ax

@@follow_near_code:
		mov	ax,[es:si]
@@follow_code:
		mov	[cur_ip],ax
		mov	word [cur_line],0
		call	update_commands
		stc
		ret

@@data_search:
		call	search_bytes
		clc
		ret

@@data_search_next:
		call	search_next
		clc
		ret

@@actions	dw	kbEsc
		dw	@@quit		- @@first_action
		dw	kbF10
		dw	@@quit		- @@first_action
		dw	kbAltX
		dw	@@quit		- @@first_action
		dw	kbEnter
		dw	@@quit		- @@first_action
		dw	kbRight
		dw	@@right		- @@first_action
		dw	kbLeft
		dw	@@left		- @@first_action
		dw	kbUp
		dw	@@up		- @@first_action
		dw	kbDown
		dw	@@down		- @@first_action
		dw	kbPgUp
		dw	@@page_up	- @@first_action
		dw	kbPgDn
		dw	@@page_down	- @@first_action
		dw	kbCtrlLeft
		dw	@@byte_left	- @@first_action
		dw	kbCtrlRight
		dw	@@byte_right	- @@first_action
		dw	kbCtrlZ
		dw	@@byte_left	- @@first_action
		dw	kbCtrlY
		dw	@@byte_left	- @@first_action
		dw	kbCtrlZGerman
		dw	@@byte_left	- @@first_action
		dw	kbCtrlYGerman
		dw	@@byte_left	- @@first_action
		dw	kbCtrlX
		dw	@@byte_right	- @@first_action
		dw	kbHome
		dw	@@home		- @@first_action
		dw	kbEnd
		dw	@@end		- @@first_action
		dw	kbTab
		dw	@@switch	- @@first_action
		dw	kbF12
		dw	@@switch	- @@first_action
		dw	kbCtrlD
		dw	@@dump_go_to	- @@first_action
		dw	kbAltF10
		dw	@@local_menu	- @@first_action
		dw	kbAltEqu
		dw	@@dump_reference - @@first_action
		dw	kbAltU
		dw	@@dump_user_scr - @@first_action
		dw	kbAltE
		dw	@@dump_esdi	- @@first_action
		dw	kbAltD
		dw	@@dump_dssi	- @@first_action
		dw	kbAltB
		dw	@@dump_esbx	- @@first_action
		dw	kbAltS
		dw	@@dump_ssbp	- @@first_action
		dw	kbAltA
		dw	@@dump_dxax	- @@first_action
		dw	kbAltF
		dw	@@dump_dsdx	- @@first_action
		dw	kbAltI
		dw	@@dump_essi	- @@first_action
		dw	kbAltJ
		dw	@@dump_dsdi	- @@first_action
		dw	kbAltC
		dw	@@dump_csip	- @@first_action
		dw	kbAltL
		dw	@@dump_cur_line	- @@first_action
		dw	kbCtrlF
		dw	@@dump_follow	- @@first_action
		dw	kbCtrlB
		dw	@@block		- @@first_action
		dw	kbCtrlS
		dw	@@search	- @@first_action
		dw	kbCtrlL
		dw	@@search_next	- @@first_action
		dw	kbAltZ
		dw	@@zoom_dump	- @@first_action
		dw	kbAltF5
		dw	@@user_scr	- @@first_action
		dw	0			;end of table

follow_data_seg		equ	@@follow_data_seg
follow_far_data		equ	@@follow_far_data
follow_near_data	equ	@@follow_near_data
follow_code_seg		equ	@@follow_code_seg
follow_far_code		equ	@@follow_far_code
follow_near_code	equ	@@follow_near_code
data_search		equ	@@data_search
data_search_next	equ	@@data_search_next

;=============================================================================
; edit_stack
;-----------------------------------------------------------------------------
edit_stack:
		or	byte [edit_mode],ED_STACK
		call	hide_focus
		xor	dx,dx
		mov	bp,edit_sssp_params
		mov	[stack_mode],bp
		call	calc_ss_addr
		call	update_stack

@@next:
		call	large_cursor
		push	dx
		add	dl,[ds:bp+0]		;X
		neg	dh
		add	dh,STK_BOT_Y
		call	gotoxy
		pop	dx

@@read_key:
		call	read_key
		mov	di,@@actions - 3
		call	action_short_jump
		jc	@@not_found
		add	di,@@first_action
		jmp	di
@@not_found:
		call	ascii_to_bin
		jc	@@next
		call	calc_ss_addr		;DI -> word
		push	ds
		mov	ds,[cpu_regs+REG_SS]
		call	set_half_byte
		pop	ds
		jmp	@@right

@@first_action:

;@@local_menu:
;		push	si
;		mov	si,dump_menu_struc
;
;@@exec_menu:
;		call	_local_menu
;		pop	si
;		jc	@@jump_to_next
;

@@update:
		call	calc_ss_addr
		call	update_dump
		call	update_commands
		call	update_stack
		jmp	@@next
; Right
@@right:
		inc	dl
		cmp	dl,3
		jbe	@@update
		mov	dl,0
; Down
@@down:
		dec	dh
		jns	@@update
		sub	word [ds:bp+1],2
		mov	dh,0
		jmp	@@update
; Left
@@left:
		dec	dl
		jns	@@update
		mov	dl,3
; Up
@@up:
		inc	dh
		cmp	dh,[stack_height]
		jb	@@update
		mov	dh,[stack_height]
		dec	dh
		add	word [ds:bp+1],2
		jmp	@@update
; Page Up
@@page_up:
		mov	al,[stack_height]
		shl	al,1
		cbw
		add	word [ds:bp+1],ax
		jmp	@@update
; Page Down
@@page_down:
		mov	al,[stack_height]
		shl	al,1
		cbw
		sub	word [ds:bp+1],ax
		jmp	@@update
; Tab
@@switch:
		xor	dl,dl
		xor	bp,word ED_SSSP ^ ED_SSBP
		mov	[stack_mode],bp
		xchg	dx,[stack_wnd_xy]
		jmp	@@update		;@@next
; ESC, Enter
@@quit:
		call	no_cursor
		and	byte [edit_mode],~ED_STACK
		mov	word [stack_wnd_xy],0
		call	draw_screen
		call	display_focus
		ret

@@actions	dw	kbEsc
		db	@@quit		- @@first_action
		dw	kbF10
		db	@@quit		- @@first_action
		dw	kbAltX
		db	@@quit		- @@first_action
		dw	kbEnter
		db	@@quit		- @@first_action
		dw	kbUp
		db	@@up		- @@first_action
		dw	kbDown
		db	@@down		- @@first_action
		dw	kbLeft
		db	@@left		- @@first_action
		dw	kbRight
		db	@@right		- @@first_action
		dw	kbPgUp
		db	@@page_up	- @@first_action
		dw	kbPgDn
		db	@@page_down	- @@first_action
		dw	kbTab
		db	@@switch	- @@first_action
		dw	kbF12
		db	@@switch	- @@first_action
		dw	0

;=============================================================================
; ---
; In:	DH -- line
; Out:	DI -> word
; Modf:	---
; Call:	---
; Use:	---
;
calc_ss_addr:
		xchg	ax,di
		mov	al,dh
		add	al,dh
		cbw
		add	ax,[ds:bp+1]
		cmp	bp,edit_sssp_params
		je	@@sssp_01
		add	ax,[cpu_regs+REG_BP]
@@sssp_01:
		mov	[ds:bp+3],ax
		xchg	ax,di
		ret

;=============================================================================
; block_menu
;-----------------------------------------------------------------------------
block_menu:
		mov	si,block_menu_struc
		call	_local_menu
		jc	@@exit
		call	ax
@@exit:
		ret
;=============================================================================
; dump_go_to
;-----------------------------------------------------------------------------
dump_go_to:
		mov	ax,16*256+40
		mov	di,dump_addr_string
		call	read_address
		jc	@@quit
		or	ch,ch
		jz	@@no_seg
		mov	[dump_seg],dx
		cmp	ch,2
		je	@@no_offs
@@no_seg:
		mov	[dump_offs],ax
@@no_offs:
		call	update_dump
		clc
@@quit:
		ret
;=============================================================================
; edit_commands
;-----------------------------------------------------------------------------
edit_commands:
		cmp	word [cur_x],0
		je	@@skip
		mov	word [cur_x],0
		call	update_commands
@@skip:
		cmp	byte [focus],1
		jae	@@skip2
		mov	byte [@@focus],1
		inc	byte [focus]
		call	update_commands
@@skip2:
		push	word [read_exit_keys]
		mov	word [read_exit_keys],asm_exit_keys
		push	cs
		pop	es
@@next_line:
		mov	si,[cur_line_ip]
		call	unassemble
		mov	di,buffer+CMD_POS+MAX_ASM_SIZE-2
		mov	cx,MAX_ASM_SIZE-2
		std
		mov	al,' '
	repe	scasb
		cld
		scasw
		mov	al,0
		stosb
		add	cx,3
		mov	si,buffer+CMD_POS
		mov	di,asm_line
	rep	movsb
;		mov	AsmLine+MaxAsmSize-1,al
@@again:
		mov	di,asm_line
		mov	ah,[cur_line]
;		add	ah,CPUy
		mov	al,CMD_POS-3
		mov	cx, ASM_LINE_SIZE-1
		mov	dx, MAX_ASM_SIZE
		mov	bx,atsAssembler*256+atAssembler
		cmp	byte [cur_line_brk],1
		jne	@@read_line
		mov	bx,atsBrkAssembler*256+atBrkAssembler
@@read_line:
		call	read_line
		cmp	ax,kbEsc
		je	@@quit
		cmp	ax,kbF10
		je	@@quit
		cmp	ax,kbAltX
		je	@@quit
		cmp	ax,kbDown
		jne	@@1
		call	cur_down
		jmp	@@next_line
@@1:		cmp	ax,kbUp
		jne	@@2
		call	cur_up
		jmp	@@next_line
@@2:		cmp	ax,kbPgDn
		jne	@@3
		call	page_down
		jmp	@@next_line
@@3:		cmp	ax,kbPgUp
		jne	@@4
		call	page_up
		jmp	@@next_line
@@4:
		mov	si,asm_line
		mov	di,[cur_line_ip]
		push	es
		mov	es,[unasm_seg]
		mov	al,[es:di]
		inc	byte [es:di]
		inc	al
		cmp	al,[es:di]
		jne	@@rom_error
		dec	byte [es:di]
		mov	bx,di
		call	assemble
		pop	es
		jnc	@@ok
		cmp	ax,ER_EMPTY
		je	@@ok_1
		dec	ax
		dec	ax
		cmp	ax,9-2
		ja	@@unk_err
		mov	bx,ax
		shl	bx,1
		mov	si,[asm_errors+bx]
		jmp	@@err_msg
@@unk_err:
		mov	si,asm_error_msg
@@err_msg:
		call	_error_message
		jmp	@@again
@@ok:
		call	update_dump
@@ok_1:
		call	cur_down
		jmp	@@next_line
@@rom_error:
		pop	es
		mov	dx,code_not_in_ram
		call	_error_message
@@quit:
		cmp	byte [@@focus],1
		jne	@@quittt
		mov	byte [@@focus],0
		dec	byte [focus]
@@quittt:
		call	update_commands
		pop	word [read_exit_keys]
		ret

@@focus		db	0


;=============================================================================
%include	"bp.inc"
;=============================================================================


;=============================================================================
; execute_menu
;-----------------------------------------------------------------------------
execute_menu:
		call	hide_focus
		mov	si,menu_structure
		call	exec_menu
		call	display_focus
		call	check_shared_area
		mov	al,[cpu_type]
		sub	al,a86
		mov	[opt_cpu],al
		cmp	byte [quit_flag],1
		jne	@@1
		jmp	quit_act
@@1:
		ret
;=============================================================================
; quit_proc
;-----------------------------------------------------------------------------
quit_proc:
		mov	byte [quit_flag],1
		ret
;=============================================================================
; set_cpu_88
;-----------------------------------------------------------------------------
set_cpu_88:
;		mov	OptCPU,0
		mov	byte [cpu_type],a86
		jmp	update_commands
;=============================================================================
; set_cpu_286
;-----------------------------------------------------------------------------
set_cpu_286:
;		mov	OptCPU,1
		mov	byte [cpu_type],a286
		jmp	update_commands
;=============================================================================
; set_cpu_386
;-----------------------------------------------------------------------------
set_cpu_386:
;		mov	OptCPU,2
		mov	byte [cpu_type],a386
		jmp	update_commands
;=============================================================================
; set_cpu_486
;-----------------------------------------------------------------------------
set_cpu_486:
;		mov	OptCPU,3
		mov	byte [cpu_type],a486
		jmp	update_commands
;=============================================================================
; set_cpu_auto
;-----------------------------------------------------------------------------
set_cpu_auto:
		mov	al,[system_cpu]
		sub	al,1
		adc	al,0
		mov	[opt_cpu],al
		add	al,a86
		mov	[cpu_type],al
		jmp	update_commands
;=============================================================================
; set_swap_off
;-----------------------------------------------------------------------------
set_swap_off:
		mov	byte [swap_mode],0
		ret
;=============================================================================
; set_swap_smart
;-----------------------------------------------------------------------------
set_swap_smart:
		mov	byte [swap_mode],1
		ret
;=============================================================================
; set_swap_on
;-----------------------------------------------------------------------------
set_swap_on:
		mov	byte [swap_mode],2
		ret
;=============================================================================
; cpu_local_menu
;-----------------------------------------------------------------------------
cpu_local_menu:
		mov	si,cpu_local_struc
		call	_local_menu
		jc	@@exit
		jmp	ax
@@exit:
		ret
;=============================================================================
; continue_init
;-----------------------------------------------------------------------------
continue_init:
		cmp	byte [loaded],1
		je	@@check_resident
		mov	si,load_error_msg
		call	_error_message

@@check_resident:
		test	byte [cmd_line_options],CMD_RESIDENT
		jz	start_key_loop
		call	_resident

start_key_loop:

		call	draw_screen
main_key_loop:
		mov	word [keys],main_loop
		call	key_loop
		cmp	byte [resident_mode],0
		jne	main_key_loop

;----- Exit from ACT ----------------------------------------------------------
quit_act:
%if _TERMINAL
	call disconnect_terminal
%endif

		call	unload_program
		call	set_my_pid
		call	restore_int_table
		call	restore_screen

		mov	al,[my_port_21]
		out	21h,al
		mov	ax,4c00h
		int	21h
;=============================================================================
; _error_message
;-----------------------------------------------------------------------------
_error_message:
		call	error_message
		jmp	check_shared_area
;=============================================================================
; _msg_box
;-----------------------------------------------------------------------------
_msg_box:
		call	message_box
		jmp	check_shared_area
;=============================================================================
; _local_menu
;-----------------------------------------------------------------------------
_local_menu:
		call	local_menu
		pushf
		call	check_shared_area
		popf
		ret

toggle_ascii:
	xor byte [cmd_line_options], CMD_ASCII
	call initialise_draw_table

;=============================================================================
; draw_screen
;-----------------------------------------------------------------------------

		; INP:	dx = data coordinates of dump cursor within
		;		the dump window (if active)
		;	dl = 0..15 = data column
		;	dh = 0 .. dump height minus one = data row
draw_screen:
	mov word [dialog_accelerators_behind_last], dialog_accelerators
			; Magic: If we draw the screen then we have finished
			;  processing any open dialog. Eg for search, pressing
			;  the Search button with an invalid search string
			;  results in re-executing the Search dialog, but
			;  without re-initialising it. Therefore this reset
			;  does not belong at the end of exec_dialog.
		call	draw_cpu_window
update_screen:
		call	update_stack
		call	update_watch_label
update_regs_and_cmds:
		call	_update_dump
		call	update_registers
%if _TERMINAL
		call	update_commands
		jmp update_terminal
%else
		jmp	update_commands
%endif

initialise_draw_table:
	call pushr
	push ds
	pop es
	mov si, drawing_list_table
	test byte [cmd_line_options], CMD_ASCII
	jz @F
	mov si, ascii_list_table
@@:
	mov cx, drawing_list_table.size
	mov di, draw_table
	rep movsb

	drawchoice '', '>'
	mov al, byte [DRAWADDRESS]
	mov si, arrow_offset_list_table
	mov cx, arrow_offset_list_table.amount
@@:
	xchg ax, di
	lodsw
	xchg ax, di
	stosb
	loop @B
	call popr
	retn

;=============================================================================
; E0F
;=============================================================================

